# Local
from .common import last_4


class AuthMethod:
    """
    An abstract class for authentication methods.

    **Design note:** we currently still implement our own interface for API
    authentication instead of making a custom ``httpx.AuthType`` (which would
    be more elegant and require less bespoke code) because some of our APIs'
    authorization methods require adding parameters to the body. Once the
    ``Request`` object has been instantiated, its body is already encoded based
    on the ``json`` keyword argument to its constructor. Therefore, it has no
    concept of "body parameters" that can be trivially updated in the scope of
    ``AuthType.auth_flow()``, unless we were to implement our own custom
    ``Request`` subclass that is aware of its body payload being a map-like
    object and rewrite the :attr:`pagerduty.ApiClient.request` method to
    prepare and send requests. Such a deep level of customization may make the
    client more brittle to upstream changes.

    :param secret:
        The API credential to be used for authentication.
    """

    def __init__(self, secret):
        self.secret = secret

    @property
    def auth_header(self) -> dict:
        """
        Generates the header that will be used for authenticating with
        the PagerDuty API
        """
        raise NotImplementedError

    @property
    def auth_param(self) -> dict:
        """
        Generates an authentication parameter for the body of the request.
        """
        raise NotImplementedError

    @property
    def secret(self):
        """
        Returns the API secret associated with the authentication method.
        """
        return self._secret

    @secret.setter
    def secret(self, secret):
        self._secret = secret

    @property
    def trunc_secret(self) -> str:
        """
        Returns a truncated version of the API credential for display purposes.
        """
        return last_4(self.secret)


class HeaderAuthMethod(AuthMethod):
    """
    Abstract base class for auth methods that authenticate using request headers.

    In this class, ``auth_param`` is defined such that it injects no parameters
    into the body of the request by default, leaving ``auth_header``
    un-implemented.
    """

    @property
    def auth_param(self) -> dict:
        return {}


class BodyParameterAuthMethod(AuthMethod):
    """
    Abstract base class for APIs that authenticate using a body parameter.

    In this class, ``auth_header`` is defined such that it adds no headers to
    the request, but it leaves ``auth_param`` un-implemented to require its
    implementation.
    """

    @property
    def auth_header(self) -> dict:
        return {}


class PassThruHeaderAuthMethod(HeaderAuthMethod):
    """
    Auth method that sets the ``Authorization`` header verbatim.

    This is for use cases where an ``Authorization`` header must be passed
    through for authentication, i.e. in an API proxy or MCP server.
    """

    @property
    def auth_header(self) -> dict:
        return {"Authorization": self.secret}



import os
import sys

UNKNOWN_VERSION = "6.*.*"


def get_version_via_tomllib():
    import tomllib

    base_project_path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..")
    )
    with open(os.path.join(base_project_path, "pyproject.toml"), "rb") as f:
        pkg_meta = tomllib.load(f)
        return pkg_meta.get("project", {}).get("version", UNKNOWN_VERSION)


def get_version():
    if sys.version_info.major == 3 and sys.version_info.minor < 8:
        # There is no way to obtain the version from the package metadata
        # because the necessary importlib features have not yet been added. At
        # some point we need to drop support for these versions. We only care
        # about Python major version 3 because version 2 is already not
        # supported and Python 4 isn't out yet.
        return UNKNOWN_VERSION
    else:
        try:
            # Use package metadata introspection to get the version:
            from importlib.metadata import version

            return version(__package__)
        except Exception:
            # No package has been built/installed yet, so this is a stopgap to
            # avoid errors in local unit tests and documentation builds:
            if sys.version_info.minor < 11:
                # tomllib was introduced in 3.11
                return UNKNOWN_VERSION
            else:
                # Use tomllib so the correct version number goes into the doc
                # build:
                return get_version_via_tomllib()


__version__ = get_version()



#!/usr/bin/env python3

import argparse
import socket
import sys
from .events_api_v2_client import EventsApiV2Client

HOSTNAME = socket.gethostname()
DEFAULT_SOURCE = f"pagerduty.cli on {HOSTNAME}"


def run(argv):
    parser = argparse.ArgumentParser(
        description="PagerDuty Events API V2 Command Line Interface"
    )
    parser.add_argument(
        "action",
        choices=["trigger", "acknowledge", "resolve"],
        help="Action to perform: trigger, acknowledge, or resolve an incident",
    )
    parser.add_argument(
        "-k",
        "--routing-key",
        required=True,
        help="PagerDuty Events API v2 routing key",
    )
    parser.add_argument(
        "-i",
        "--dedup-key",
        default=None,
        help="Deduplication key for the incident",
    )
    parser.add_argument(
        "--description", help="Summary/description of the alert"
    )
    parser.add_argument(
        "--source", help="Source of the alert", default=DEFAULT_SOURCE
    )

    args = parser.parse_args(argv)

    # Create the Events API client
    client = EventsApiV2Client(args.routing_key)

    try:
        # Handle the different actions
        if args.action == "trigger":
            if not args.description:
                parser.error("--description is required for trigger action")
            dedup_key = client.trigger(
                args.description, args.source, dedup_key=args.dedup_key
            )
            print(
                f"Alert triggered successfully. Deduplication key: {dedup_key}"
            )

        elif args.action == "acknowledge":
            if not args.dedup_key:
                parser.error(
                    "-i/--dedup-key is required for acknowledge action"
                )

            dedup_key = client.acknowledge(args.dedup_key)
            print(
                "Alert acknowledged successfully. Deduplication key: "
                + dedup_key
            )

        elif args.action == "resolve":
            if not args.dedup_key:
                parser.error("-i/--dedup-key is required for resolve action")

            dedup_key = client.resolve(args.dedup_key)
            print(
                f"Alert resolved successfully. Deduplication key: {dedup_key}"
            )

    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


def main():
    run(sys.argv[1:])


if __name__ == "__main__":
    main()



# Core
from copy import deepcopy
from datetime import datetime
from typing import List, Optional

# PyPI

# Local
from .api_client import ApiClient
from .auth_method import BodyParameterAuthMethod
from .common import successful_response, try_decoding, truncate_text
from .errors import ServerHttpError


class RoutingKeyAuthMethod(BodyParameterAuthMethod):
    """
    AuthMethod for Events API V2.

    Requires a ``routing_key`` parameter be set in the request body.
    """

    @property
    def auth_param(self) -> dict:
        return {"routing_key": self.secret}


class EventsApiV2Client(ApiClient):
    """
    Client class for submitting events to the PagerDuty v2 Events API.

    Implements methods for submitting events to PagerDuty through the Events
    API, including change events, and inherits from
    :class:`pagerduty.ApiClient`.  For more details on usage of this API, refer
    to the `Events API v2 documentation
    <https://developer.pagerduty.com/docs/events-api-v2/overview/>`_

    :param routing_key:
        The routing key to use for authentication with the Events API.
        Sometimes called an ``integration_key`` or ``service_key`` in legacy
        integrations.

    :param debug:
        Sets :attr:`pagerduty.ApiClient.print_debug`. Set to ``True`` to enable
        verbose command line output.
    """

    _url = "https://events.pagerduty.com"

    permitted_methods = ("POST",)

    def __init__(self, routing_key: str, debug: bool = False, **kw):
        auth_method = RoutingKeyAuthMethod(routing_key)
        super(EventsApiV2Client, self).__init__(auth_method, debug=debug, **kw)
        # See: https://developer.pagerduty.com/docs/3d063fd4814a6-events-api-v2-overview#response-codes--retry-logic
        self.retry[500] = 2  # internal server error
        self.retry[502] = 4  # bad gateway
        self.retry[503] = 6  # service unavailable
        self.retry[504] = 6  # gateway timeout

    def acknowledge(self, dedup_key: str) -> str:
        """
        Acknowledge an alert via Events API.

        :param dedup_key:
            The deduplication key of the alert to set as acknowledged.
        :returns:
            The deduplication key
        """
        return self.send_event("acknowledge", dedup_key=dedup_key)

    @property
    def event_timestamp(self) -> str:
        return datetime.utcnow().isoformat() + "Z"

    def resolve(self, dedup_key: str) -> str:
        """
        Resolve an alert via Events API.

        :param dedup_key:
            The deduplication key of the alert to resolve.
        """
        return self.send_event("resolve", dedup_key=dedup_key)

    def send_change_event(
        self,
        payload: Optional[dict] = None,
        links: Optional[List[dict]] = None,
        images: Optional[List[dict]] = None,
    ):
        """
        Send a change event to the v2 Change Events API.

        See: https://developer.pagerduty.com/docs/events-api-v2/send-change-events/

        :param payload:
            A dictionary object with keys ``summary``, ``source``,
            ``timestamp`` and ``custom_details`` as described in the above
            documentation.
        :param links:
            A list of dictionary objects each with keys ``href`` and ``text``
            representing the target and display text of each link
        :param images:
            Optional list of images to attach to the change event.
        """
        if payload is None:
            payload = {}
        if links is None:
            links = []
        if images is None:
            images = []
        event = {"payload": deepcopy(payload)}
        if links:
            event["links"] = deepcopy(links)
        if images:
            event["images"] = deepcopy(images)
        successful_response(
            self.post("/v2/change/enqueue", json=event),
            context="submitting change event",
        )

    def send_event(
        self, action: str, dedup_key: Optional[str] = None, **properties
    ) -> str:
        """
        Send an event to the v2 Events API.

        See: https://v2.developer.pagerduty.com/docs/send-an-event-events-api-v2

        :param action:
            The action to perform through the Events API: trigger, acknowledge
            or resolve.
        :param dedup_key:
            The deduplication key; used for determining event uniqueness and
            associating actions with existing incidents.
        :param **properties:
            Additional properties to set, i.e. if ``action`` is ``trigger``
            this would include ``payload``.
        :type action: str
        :type dedup_key: str
        :returns:
            The deduplication key of the incident
        """

        actions = ("trigger", "acknowledge", "resolve")
        if action not in actions:
            raise ValueError(
                "Event action must be one of: " + ", ".join(actions)
            )

        event = {"event_action": action}

        event.update(properties)
        if isinstance(dedup_key, str):
            event["dedup_key"] = dedup_key
        elif not action == "trigger":
            raise ValueError(
                "The dedup_key property is required for"
                "event_action=%s events, and it must be a string." % action
            )
        response = successful_response(
            self.post("/v2/enqueue", json=event),
            context="submitting an event to the events API",
        )
        response_body = try_decoding(response)
        if type(response_body) is not dict or "dedup_key" not in response_body:
            err_msg = (
                "Malformed response body from the events API; it is "
                'not a dict that has a key named "dedup_key" after '
                "decoding. Body = " + truncate_text(response.text)
            )
            raise ServerHttpError(err_msg, response)
        return response_body["dedup_key"]

    def submit(
        self,
        summary: str,
        source: Optional[str] = None,
        custom_details: Optional[dict] = None,
        links: Optional[List[dict]] = None,
        timestamp: Optional[str] = None,
    ):
        """
        Submit a change event.

        See: https://developer.pagerduty.com/docs/send-change-event

        This is a wrapper method for :attr:`send_change_event` that composes an
        event payload from keyword arguments and an auto-generated event
        timestamp. To send an event with a wholly custom payload, use
        :attr:`send_change_event` instead.

        :param summary:
            Summary / brief description of the change, for ``payload.summary``.
        :param source:
            A human-readable name identifying the source of the change, for the
            ``payload.source`` event property.
        :param custom_details:
            A dictionary object to use as property ``payload.custom_details``
        :param links:
            A list of dict objects to use as property ``links`` in the event.
        :param timestamp:
            Specifies an event timestamp. Must be an ISO8601-format date/time.
        :type summary: str
        :type source: str
        :type custom_details: dict
        :type links: list
        :type timestamp: str
        """
        local_var = locals()["custom_details"]
        if not (local_var is None or isinstance(local_var, dict)):
            raise ValueError("custom_details must be a dict")
        if timestamp is None:
            timestamp = self.event_timestamp
        event = {
            "payload": {
                "summary": summary,
                "timestamp": timestamp,
            }
        }
        if isinstance(source, str):
            event["payload"]["source"] = source
        if isinstance(custom_details, dict):
            event["payload"]["custom_details"] = custom_details
        if links:
            event["links"] = links
        self.send_change_event(**event)

    def trigger(
        self,
        summary: str,
        source: str,
        dedup_key: Optional[str] = None,
        severity: str = "critical",
        payload: Optional[str] = None,
        custom_details: Optional[dict] = None,
        images: Optional[List[dict]] = None,
        links: Optional[List[dict]] = None,
    ) -> str:
        """
        Send an alert-triggering event

        :param summary:
            Summary / brief description of what is wrong.
        :param source:
            A human-readable name identifying the system that is affected.
        :param dedup_key:
            The deduplication key; used for determining event uniqueness and
            associating actions with existing incidents.
        :param severity:
            Alert severity. Sets the ``payload.severity`` property.
        :param payload:
            Set the payload directly. Can be used in conjunction with other
            parameters that also set payload properties; these properties will
            be merged into the default payload, and any properties in this
            parameter will take precedence except with regard to
            ``custom_details``.
        :param custom_details:
            The ``payload.custom_details`` property of the payload. Will
            override the property set in the ``payload`` parameter if given.
        :param images:
            Set the ``images`` property of the event.
        :param links:
            Set the ``links`` property of the event.
        :type action: str
        :type custom_details: dict
        :type dedup_key: str
        :type images: list
        :type links: list
        :type payload: dict
        :type severity: str
        :type source: str
        :type summary: str
        :returns:
            The deduplication key of the incident, if any.
        """
        for local in ("payload", "custom_details"):
            local_var = locals()[local]
            if not (local_var is None or type(local_var) is dict):
                raise ValueError(local + " must be a dict")
        event = {
            "payload": {
                "summary": summary,
                "source": source,
                "severity": severity,
            }
        }
        if type(payload) is dict:
            event["payload"].update(payload)
        if type(custom_details) is dict:
            details = event.setdefault("payload", {}).get("custom_details", {})
            details.update(custom_details)
            event["payload"]["custom_details"] = details
        if images:
            event["images"] = images
        if links:
            event["links"] = links
        return self.send_event("trigger", dedup_key=dedup_key, **event)



from .version import __version__

from .common import (
    TEXT_LEN_LIMIT,
    TIMEOUT,
    deprecated_kwarg,
    http_error_message,
    last_4,
    normalize_url,
    plural_name,
    requires_success,
    singular_name,
    successful_response,
    truncate_text,
    try_decoding,
)

from .api_client import ApiClient

from .auth_method import PassThruHeaderAuthMethod

from .errors import Error, HttpError, ServerHttpError, UrlError

from .events_api_v2_client import EventsApiV2Client, RoutingKeyAuthMethod

from .jira_cloud_integration_api_client import JiraCloudIntegrationApiClient
from .jira_server_integration_api_client import JiraServerIntegrationApiClient
from .mcp_api_client import McpApiClient
from .ms_teams_integration_api_client import MsTeamsIntegrationApiClient
from .oauth_token_client import OAuthTokenClient

from .rest_api_v2_base_client import (
    ITERATION_LIMIT,
    OAuthTokenAuthMethod,
    RestApiV2BaseClient,
    TokenAuthMethod,
    auto_json,
    endpoint_matches,
    infer_entity_wrapper,
    is_path_param,
    resource_url,
    unwrap,
    wrapped_entities,
)

from .rest_api_v2_client import (
    CANONICAL_PATHS,
    CURSOR_BASED_PAGINATION_PATHS,
    ENTITY_WRAPPER_CONFIG,
    RestApiV2Client,
    canonical_path,
    entity_wrappers,
)

from .scim_api_client import ScimApiClient
from .slack_integration_api_client import SlackIntegrationApiClient
from .slack_integration_connections_api_client import (
    SlackIntegrationConnectionsApiClient,
)

# For backwards compatibility, __all__ currently includes all of the above.
# This should eventually be cleaned up so that it includes only the most-used
# interfaces, i.e.  client classes. That could be done as a minor breaking
# change in a new future major version, i.e. it would break "from pagerduty
# import *" use cases, which would be an opportunity to lessen the impact of a
# future breaking change that ends the practice of importing the helper methods
# and module configuration globals.
__all__ = [
    "__version__",
    "ApiClient",
    "CANONICAL_PATHS",
    "CURSOR_BASED_PAGINATION_PATHS",
    "ENTITY_WRAPPER_CONFIG",
    "Error",
    "EventsApiV2Client",
    "HttpError",
    "ITERATION_LIMIT",
    "JiraCloudIntegrationApiClient",
    "JiraServerIntegrationApiClient",
    "McpApiClient",
    "MsTeamsIntegrationApiClient",
    "OAuthTokenAuthMethod",
    "OAuthTokenClient",
    "PassThruHeaderAuthMethod",
    "RestApiV2BaseClient",
    "RestApiV2Client",
    "RoutingKeyAuthMethod",
    "ScimApiClient",
    "ServerHttpError",
    "SlackIntegrationApiClient",
    "SlackIntegrationConnectionsApiClient",
    "TEXT_LEN_LIMIT",
    "TIMEOUT",
    "TokenAuthMethod",
    "UrlError",
    "auto_json",
    "canonical_path",
    "deprecated_kwarg",
    "endpoint_matches",
    "entity_wrappers",
    "http_error_message",
    "infer_entity_wrapper",
    "is_path_param",
    "last_4",
    "normalize_url",
    "plural_name",
    "requires_success",
    "resource_url",
    "singular_name",
    "successful_response",
    "truncate_text",
    "try_decoding",
    "unwrap",
    "wrapped_entities",
]



from typing import List, Optional

from .api_client import ApiClient
from .common import successful_response, try_decoding


class ScimApiClient(ApiClient):
    """
    Client class for the PagerDuty SCIM API.

    This class features a method :attr:`list_users` that provides an interface
    for retrieving all users that match a given filter.

    Other endpoints of the PagerDuty SCIM API can be accessed through the
    standard methods named after the HTTP methods.

    Usage example:

    .. code-block:: python

        # Import and use OAuthTokenAuthMethod instead of TokenAuthMethod to use
        # an application OAuth token:
        from pagerduty import ScimApiClient, TokenAuthMethod

        # Instantiate:
        client = ScimApiClient(TokenAuthMethod(API_KEY))

        scim_users = client.list_users()

        create_users_response = client.post(
            '/Users',
            json = {
                ...
            }
        )
        if create_users_response.status == 201:
            print("User successfully created")

        scim_user_dict = create_users_response.json()

    """

    _url = "https://api.pagerduty.com/scim/v2"

    permitted_methods = ("DELETE", "GET", "PATCH", "POST", "PUT")

    def list_users(
        self,
        fltr: Optional[str] = None,
        start_index: int = 1,
        page_size: int = 100,
    ) -> List[dict]:
        """
        List all users using SCIM API with automatic pagination.

        :param fltr:
            Optional SCIM filter expression to limit results
        :param start_index:
            The 1-based index of the first result to return (SCIM standard)
        :param page_size:
            Number of results per page (default 100)
        :returns:
            List of all user entries from the SCIM Users endpoint
        """
        all_users = []
        current_start_index = start_index

        while True:
            params = {"startIndex": current_start_index, "count": page_size}

            if fltr:
                params["filter"] = fltr

            response = successful_response(
                self.get("/Users", params=params),
                context="SCIM list users pagination",
            )

            body = try_decoding(response)

            # Extract users from the SCIM response
            users = body.get("Resources", [])
            all_users.extend(users)

            # Check if there are more results
            total_results = body.get("totalResults", 0)
            items_per_page = body.get("itemsPerPage", len(users))

            # If no results are left, break:
            next_start_index = current_start_index + items_per_page - 1
            if next_start_index >= total_results or len(users) == 0:
                break

            # Move to next page
            current_start_index += items_per_page

        return all_users



from typing import List

from .rest_api_v2_base_client import CanonicalPath, RestApiV2BaseClient

CANONICAL_PATHS = ["/incidents/{incident_id}/meeting"]

ENTITY_WRAPPER_CONFIG = {
    # The "create a meeting" endpoint follows classic conventions.
    #
    # This dictionary was intentionally created and left empty so that it is fewer steps
    # to support antipatterns if they get added to this API in the future.
}


class MsTeamsIntegrationApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Microsoft Teams Integration API.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for the `PagerDuty MS Teams
    Integration API
    <https://developer.pagerduty.com/api-reference/2a7de89c77dc8-pager-duty-ms-teams-integration-api>`_.

    Its documentation indicates that it only supports the original "Token"
    style authentication and does not support OAuth token ("Bearer")
    authentication. For that reason, its constructor does not accept an
    "auth_type" argument, and it is assumed that the provided API key was
    generated through the PagerDuty web UI.

    For constructor arguments, see :class:`pagerduty.ApiClient`.
    """

    permitted_methods = ("POST",)

    url = "https://api.pagerduty.com/integration-ms-teams"

    def __init__(self, api_key: str, debug: bool = False, **kw):
        super(MsTeamsIntegrationApiClient, self).__init__(
            api_key, auth_type="token", debug=debug, **kw
        )
        self.headers.update(
            {
                "Accept": "application/json",
            }
        )

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG



# Core
import logging
import sys
import time

from copy import deepcopy
from random import random
from typing import Optional, Union

# PyPI
from httpx import __version__ as HTTPX_VERSION
from httpx import Client, Headers, TransportError, Response

# Local
from .auth_method import AuthMethod
from .version import __version__
from .errors import Error, HttpError, ServerHttpError, UrlError
from .common import TIMEOUT, normalize_url


class ApiClient(Client):
    """
    Base class for making HTTP requests to PagerDuty APIs

    This is an opinionated wrapper of `httpx.Client`_, with a few additional
    features:

    - The client will reattempt the request with auto-increasing cooldown/retry
      intervals, with attempt limits configurable through the :attr:`retry`
      attribute.
    - When making requests, headers specified ad-hoc in calls to HTTP verb
      functions will not replace, but will be merged into, default headers.
    - The request URL, if it doesn't already start with the REST API base URL,
      will be prepended with the default REST API base URL.
    - It will only perform requests with methods as given in the
      :attr:`permitted_methods` list, and will raise :class:`Error` for
      any other HTTP methods.

    :param auth_method:
        The authentication method to use for API requests, should be an
        instance of the AuthMethod class.
    :param debug:
        Sets :attr:`print_debug`. Set to ``True`` to enable verbose command
        line output.
    """

    log = None
    """
    A ``logging.Logger`` object for logging messages. By default it is
    configured without any handlers and so no messages will be emitted. See:
    `Logger Objects
    <https://docs.python.org/3/library/logging.html#logger-objects>`_.
    """

    max_http_attempts = 10
    """
    The number of times that the client will retry after error statuses, for
    any that are defined greater than zero in :attr:`retry`.
    """

    max_network_attempts = 3
    """
    The number of times that connecting to the API will be attempted before
    treating the failure as non-transient; a :class:`Error` exception
    will be raised if this happens.
    """

    parent = None
    """The ``super`` object (`httpx.Client`_)"""

    permitted_methods = ()
    """
    A tuple of the methods permitted by the API which the client implements.

    For instance:

    * The REST API accepts GET, POST, PUT and DELETE.
    * The Events API and Change Events APIs only accept POST.
    """

    retry = {}
    """
    A dict defining the retry behavior for each HTTP response status code.

    Each key in this dictionary is an int representing a HTTP response code.
    The behavior is specified by the int value at each key as follows:

    * ``-1`` to retry without limit.
    * ``0`` has no effect; the default behavior will take effect.
    * ``n``, where ``n > 0``, to retry ``n`` times (or up
      to :attr:`max_http_attempts` total for all statuses, whichever is
      encountered first), and then return the final response.

    The default behavior is to retry without limit on status 429, raise an
    exception on a 401, and return the `httpx.Response`_ object in any other
    case (assuming a HTTP response was received from the server).
    """

    sleep_timer = 1.5
    """
    Default initial cooldown time factor for rate limiting and network errors.

    Each time that the request makes a followup request, there will be a delay
    in seconds equal to this number times :attr:`sleep_timer_base` to the power
    of how many attempts have already been made so far, unless
    :attr:`stagger_cooldown` is nonzero.
    """

    sleep_timer_base = 2
    """
    After each retry, the time to sleep before reattempting the API connection
    and request will increase by a factor of this amount.
    """

    timeout = TIMEOUT
    """
    This is the value sent to `HTTPX`_ as the ``timeout`` parameter that
    determines the TCP read timeout.
    """

    def __init__(self, auth_method: AuthMethod, debug=False, **kw):
        self.parent = super(ApiClient, self)
        self.parent.__init__(**kw)
        self.auth_method = auth_method
        self.log = logging.getLogger(__name__)
        self.print_debug = debug
        self.retry = {}

    def after_set_auth_method(self):
        """
        Setter hook for setting or updating the authentication method.
        Child classes should implement this to perform additional steps.
        """
        pass

    @property
    def auth_method(self) -> AuthMethod:
        """
        Property representing the authentication method used for API requests.
        """
        return self._auth_method

    @auth_method.setter
    def auth_method(self, auth_method: AuthMethod):
        if not (isinstance(auth_method, AuthMethod)):
            raise ValueError(
                "auth_method must be an instance of the AuthMethod class"
            )

        self._auth_method = auth_method
        self.after_set_auth_method()

    def cooldown_factor(self) -> float:
        return self.sleep_timer_base * (1 + self.stagger_cooldown * random())

    def normalize_params(self, params: dict) -> dict:
        """
        Modify the user-supplied parameters to ease implementation

        :returns:
            The query parameters after modification
        """
        return params

    def normalize_url(self, url: str) -> str:
        """Compose the URL whether it is a path or an already-complete URL"""
        return normalize_url(self.url, url)

    def postprocess(self, response: Response):
        """
        Perform supplemental actions immediately after receiving a response.

        This method is called once per request not including retries, and can be
        extended in child classes.
        """
        pass

    def prepare_headers(
        self, method: str, user_headers: Optional[dict] = None
    ) -> dict:
        """
        Append all necessary headers per-request.

        The upstream client class `httpx.Client`_ will merge headers into the
        default instance headers, so any defaults set by the end user in the
        mutable ``headers`` property will already be merged in, and the headers
        specified at request time will take precendence.

        This method just merges in the PagerDuty API client's defaults on a
        per-request basis to avoid storing per-API header settings like API
        keys in drift-able mutable stateful instance attributes, and to enforce
        the correct headers on each request, especially where they might differ
        per-API / per request.

        :param method:
            The HTTP method, in upper case.
        :param user_headers:
            Headers that can be specified to override default values.
        :returns:
            The final list of headers to use in the request
        """
        headers = Headers({})
        # Override the default user-agent with the per-class user_agent attr:
        headers["User-Agent"] = self.user_agent
        # A nearly universal convention: whenever sending a POST, PUT or PATCH,
        # the Content-Type header must be "application/json":
        if method in ("POST", "PUT", "PATCH"):
            headers["Content-Type"] = "application/json"
        # Add headers passed in per-request as an additional argument, letting
        # them take precedence over the defaults:
        if type(user_headers) is dict:
            headers.update(user_headers)
        # Add authentication header:
        headers.update(self.auth_method.auth_header)
        return headers

    @property
    def print_debug(self) -> bool:
        """
        Printing debug flag

        If set to True, the logging level of :attr:`log` is set to
        ``logging.DEBUG`` and all log messages are emitted to ``sys.stderr``.
        If set to False, the logging level of :attr:`log` is set to
        ``logging.NOTSET`` and the debugging log handler that prints messages
        to ``sys.stderr`` is removed. This value thus can be toggled to enable
        and disable verbose command line output.

        It is ``False`` by default and it is recommended to keep it that way in
        production settings.
        """
        return self._debug

    @print_debug.setter
    def print_debug(self, debug: bool):
        self._debug = debug
        if debug and not hasattr(self, "_debugHandler"):
            self.log.setLevel(logging.DEBUG)
            self._debugHandler = logging.StreamHandler()
            self.log.addHandler(self._debugHandler)
        elif not debug and hasattr(self, "_debugHandler"):
            self.log.setLevel(logging.NOTSET)
            self.log.removeHandler(self._debugHandler)
            delattr(self, "_debugHandler")
        # else: no-op; only happens if debug is set to the same value twice

    def request(self, method: str, url: str, **kwargs) -> Response:
        """
        Make a generic PagerDuty API request.

        :param method:
            The request method to use. Case-insensitive. May be one of get, put,
            post or delete.
        :param url:
            The path/URL to request. If it does not start with the base URL,
            the base URL will be prepended.
        :param **kwargs:
            Custom keyword arguments to pass to ``httpx.Client.request``.
        :type method: str
        :type url: str
        :returns:
            The `httpx.Response`_ object corresponding to the HTTP response
        """
        sleep_timer = self.sleep_timer
        network_attempts = 0
        http_attempts = {}
        method = method.strip().upper()
        if method not in self.permitted_methods:
            m_str = ", ".join(self.permitted_methods)
            raise Error(
                f"Method {method} not supported by this API. Permitted "
                f"methods: {m_str}"
            )
        req_kw = deepcopy(kwargs)
        full_url = self.normalize_url(url)
        endpoint = "%s %s" % (method.upper(), full_url)

        # Add in any headers specified in keyword arguments:
        headers = kwargs.get("headers", {})
        # Add some defaults:
        req_kw.update(
            {
                "headers": self.prepare_headers(method, user_headers=headers),
                "timeout": self.timeout,
                "auth": None,
                "follow_redirects": False,
                "cookies": None,
            }
        )

        # Add authentication parameter, if the API requires it and it is a
        # request type that includes a body:
        if method in ("POST", "PUT", "PATCH"):
            for body_key in ("json", "data"):
                if body_key in req_kw and type(req_kw[body_key]) is dict:
                    req_kw[body_key].update(self.auth_method.auth_param)

        # Special changes to user-supplied query parameters, for convenience:
        if "params" in kwargs and kwargs["params"]:
            req_kw["params"] = self.normalize_params(kwargs["params"])

        # Make the request (and repeat w/cooldown if the rate limit is reached):
        while True:
            try:
                response = self.parent.request(method, full_url, **req_kw)
                self.postprocess(response)
            except TransportError as e:
                network_attempts += 1
                if network_attempts > self.max_network_attempts:
                    error_msg = (
                        f"{endpoint}: Non-transient network "
                        "error; exceeded maximum number of attempts "
                        f"({self.max_network_attempts}) to connect to the API."
                    )
                    raise Error(error_msg) from e
                sleep_timer *= self.cooldown_factor()
                self.log.warning(
                    "%s: HTTP or network error: %s. retrying in %g seconds.",
                    endpoint,
                    e.__class__.__name__,
                    sleep_timer,
                )
                time.sleep(sleep_timer)
                continue

            status = response.status_code
            retry_logic = self.retry.get(status, 0)
            if status // 100 == 3:
                # Redirects are not expected but if they happen, fail noisily:
                raise ServerHttpError(
                    f"Received status {status} in response to {method} "
                    f"{full_url}, but PagerDuty APIs are not expected to "
                    "issue redirects.",
                    response,
                )
            elif not response.is_success and retry_logic != 0:
                # Take special action as defined by the retry logic
                if retry_logic != -1:
                    # Retry a specific number of times (-1 implies infinite)
                    if (
                        http_attempts.get(status, 0) >= retry_logic
                        or sum(http_attempts.values()) > self.max_http_attempts
                    ):
                        lower_limit = retry_logic
                        if lower_limit > self.max_http_attempts:
                            lower_limit = self.max_http_attempts
                        self.log.error(
                            "%s: Non-transient HTTP error: exceeded "
                            "maximum number of attempts (%d) to make a "
                            "successful request. Currently encountering "
                            "status %d.",
                            endpoint,
                            lower_limit,
                            status,
                        )
                        return response
                    http_attempts[status] = 1 + http_attempts.get(status, 0)
                sleep_timer *= self.cooldown_factor()
                self.log.warning(
                    "%s: HTTP error (%d); retrying in %g seconds.",
                    endpoint,
                    status,
                    sleep_timer,
                )
                time.sleep(sleep_timer)
                continue
            elif status == 429:
                sleep_timer *= self.cooldown_factor()
                self.log.debug(
                    "%s: Hit API rate limit (status 429); retrying in %g "
                    "seconds",
                    endpoint,
                    sleep_timer,
                )
                time.sleep(sleep_timer)
                continue
            elif status == 401:
                # Stop. Authentication failed. We shouldn't try doing any more,
                # because we'll run into the same problem later anyway.
                raise HttpError(
                    "Received 401 Unauthorized response from the API. The API "
                    f"credential ({self.trunc_key}) may be invalid or "
                    "deactivated, or the client is configured for the wrong "
                    "service region (in which case updating the url property "
                    "accordingly may resolve the issue).",
                    response,
                )
            else:
                # All went according to plan.
                return response

    @property
    def stagger_cooldown(self) -> float:
        """
        Randomizing factor for wait times between retries during rate limiting.

        If set to number greater than 0, the sleep time for rate limiting will
        (for each successive sleep) be adjusted by a factor of one plus a
        uniformly-distributed random number between 0 and 1 times this number,
        on top of the base sleep timer :attr:`sleep_timer_base`.

        For example:

        * If this is 1, and :attr:`sleep_timer_base` is 2 (default), then after
          each status 429 response, the sleep time will change overall by a
          random factor between 2 and 4, whereas if it is zero, it will change
          by a factor of 2.
        * If :attr:`sleep_timer_base` is 1, then the cooldown time will be
          adjusted by a random factor between one and one plus this number.

        If the number is set to zero, then this behavior is effectively
        disabled, and the cooldown factor (by which the sleep time is adjusted)
        will just be :attr:`sleep_timer_base`

        Setting this to a nonzero number helps avoid the "thundering herd"
        effect that can potentially be caused by many API clients making
        simultaneous concurrent API requests and consequently waiting for the
        same amount of time before retrying.  It is currently zero by default
        for consistent behavior with previous versions.
        """
        if hasattr(self, "_stagger_cooldown"):
            return self._stagger_cooldown
        else:
            return 0

    @stagger_cooldown.setter
    def stagger_cooldown(self, val: Union[float, int]):
        if type(val) not in [float, int] or val < 0:
            raise ValueError(
                "Cooldown randomization factor stagger_cooldown "
                "must be a positive real number"
            )
        self._stagger_cooldown = val

    @property
    def trunc_key(self) -> str:
        """Truncated key for secure display/identification purposes."""
        return self.auth_method.trunc_secret

    @property
    def url(self) -> str:
        """
        The base URL for the API being called.

        Must be a HTTPS URL. For REST API v2 in the US service region, this is
        ``https://api.pagerduty.com``; for Events API v2 it is
        ``https://events.pagerduty.com``.  This property must be set when using
        a service region other than US Production.
        """
        return self._url

    @url.setter
    def url(self, new_base_url: str):
        if not new_base_url.startswith("https://"):
            raise UrlError("API base URL must use scheme https://")
        self._url = new_base_url

    @property
    def user_agent(self) -> str:
        return "python-pagerduty/%s python-httpx/%s Python/%d.%d" % (
            __version__,
            HTTPX_VERSION,
            sys.version_info.major,
            sys.version_info.minor,
        )



from typing import List

from .rest_api_v2_base_client import CanonicalPath, RestApiV2BaseClient

CANONICAL_PATHS = ["/accounts_mappings", "/accounts_mappings/{id}"]

ENTITY_WRAPPER_CONFIG = {"GET /accounts_mappings/{id}": None}


class JiraCloudIntegrationApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Jira Server Integration API.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for the `PagerDuty Jira Cloud
    Integration API
    <https://developer.pagerduty.com/api-reference/70ea43d07719f-pager-duty-jira-cloud-integration-api>`_.

    For constructor arguments, see :class:`pagerduty.RestApiV2BaseClient`.
    """

    _url = "https://api.pagerduty.com/integration-jira-cloud"

    permitted_methods = ("GET",)

    def __init__(
        self, api_key: str, auth_type: str = "token", debug: bool = False, **kw
    ):
        super(JiraCloudIntegrationApiClient, self).__init__(
            api_key, auth_type=auth_type, debug=debug, **kw
        )
        self.headers.update(
            {
                "Accept": "application/json",
                # All requests in the reference and not just data-bearing create/update
                # methods have this header, so it should also be included in GET:
                "Content-Type": "application/json",
            }
        )

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG



import urllib.parse
import base64
import hashlib
import secrets
from typing import Optional, Tuple

from copy import deepcopy

from httpx import Response

from .api_client import ApiClient
from .auth_method import BodyParameterAuthMethod
from .common import (
    datetime_to_relative_seconds,
    relative_seconds_to_datetime,
    successful_response,
    try_decoding,
)
from .errors import ServerHttpError
from .rest_api_v2_client import RestApiV2Client


class ClientCredentialsAuthMethod(BodyParameterAuthMethod):
    """
    AuthMethod used in OAuth token exchange requests.

    This AuthMethod requires the client secret and client ID, which are then
    transmitted to identity.pagerduty.com as parameters in the HTML
    form-encoded body.

    :param client_secret:
        The client secret used for creating OAuth tokens for the application
    :param client_id:
        The client ID of the provisioned OAuth application
    """

    def __init__(self, client_secret, client_id):
        self.secret = client_secret
        self.client_id = client_id  # Additional required parameter

    @property
    def auth_param(self) -> dict:
        return {"client_id": self.client_id, "client_secret": self.secret}


class OAuthTokenClient(ApiClient):
    """
    Client with helpers for performing an OAuth exchange for an access token.

    Tokens obtained using the client can then be used to authenticate REST API
    clients, e.g.

    .. code-block:: python

        oauth_exchange_client = pagerduty.OAuthTokenClient(
            client_secret,
            client_id
        )
        oauth_response = oauth_exchange_client.get_scoped_app_token('read')
        client = pagerduty.RestApiV2Client(
            oauth_response['access_token'],
            auth_type='bearer'
        )

    Requires `registering a PagerDuty App
    <https://developer.pagerduty.com/docs/register-an-app>`_ to obtain the
    necessary credentials, and must be used in the context of an OAuth2
    authorization flow.

    For further details, refer to:

    - `OAuth Functionality <https://developer.pagerduty.com/docs/oauth-functionality>`_
    - `User OAuth Token via Code Grant <https://developer.pagerduty.com/docs/user-oauth-token-via-code-grant>`_
    - `Obtaining an App OAuth Token <https://developer.pagerduty.com/docs/app-oauth-token>`_

    :param client_secret:
        The secret associated with the application.
    :param client_id:
        The client ID provided when registering the application.
    :param debug:
        Passed to the parent constructor as the ``debug`` argument. See:
        :class:`ApiClient`
    """

    permitted_methods = ("POST",)

    url = "https://identity.pagerduty.com"

    early_refresh_buffer = 86400
    """
    Number of seconds before the expiration date to perform a token refresh.

    Used when constructing a :class:`RestApiV2Client`; if the current date/time
    is after the expiration date, or less than this number of seconds before
    it, a token refresh is first performed.

    If the application is expected to use the resulting client object for more
    than this amount of time between each new call to :attr:`refresh_client`,
    this value can be set higher.

    By default, this is 24 hours.
    """

    def __init__(self, client_secret: str, client_id: str, debug=False, **kw):
        """
        Create an OAuth token client
        """
        auth_method = ClientCredentialsAuthMethod(client_secret, client_id)

        super(OAuthTokenClient, self).__init__(auth_method, debug=debug, **kw)

    def amended_auth_response(self, response: Response) -> dict:
        """
        Amends the auth response map as necessary for other functionality.

        :param auth_response:
            A response from the /oauth/token endpoint as a dictionary.
        :returns:
            The same response from the API in dictionary form with an
            additional key ``expiration_date`` containing the expiration
            date/time of the token in ISO8601 format. This value can then be
            used in :attr:`refresh_client`.
        """
        response_json = try_decoding(successful_response(response))
        if "expires_in" not in response_json:
            raise ServerHttpError(
                'Auth response did not include expected key "expires_in".',
                response,
            )
        response_json.update(
            {
                "expiration_date": relative_seconds_to_datetime(
                    response_json["expires_in"]
                )
            }
        )
        return response_json

    def authorize_url(self, scope: str, redirect_uri: str) -> str:
        """
        The URL in PagerDuty that a user will visit to authorize an app.

        :param scope:
            Scope of the OAuth grant requested
        :param redirect_uri:
            The redirect URI in the application that receives the authorization
            code through client redirect from PagerDuty
        :returns:
            The formatted authorize URL.
        """
        return self.get_authorize_url(
            self.auth_method.client_id, scope, redirect_uri
        )

    @classmethod
    def get_authorize_url(
        cls, client_id: str, scope: str, redirect_uri: str
    ) -> str:
        """
        Generate an authorize URL.

        This method can be called directly to circumvent the need to produce a
        client secret that is otherwise required to instantiate an object.

        This is the URL that the user initially visits in the application to
        authorize the application, which will ultimately redirect the user to
        ``redirect_uri`` but with the authorization code.

        :param client_id:
            Client ID of the application
        :param scope:
            Scope of the OAuth grant requested
        :param redirect_uri:
            The redirect URI in the application that receives the authorization
            code through client redirect from PagerDuty
        :returns:
            The formatted authorize URL.
        """
        return (
            cls.url
            + "/oauth/authorize?"
            + urllib.parse.urlencode(
                [
                    ("client_id", client_id),
                    ("redirect_uri", redirect_uri),
                    ("response_type", "code"),
                    ("scope", scope),
                ]
            )
        )

    def get_new_token(self, **kw) -> dict:
        """
        Make a token request.

        There should not be any need to call this method directly. Each of the
        supported types of token exchange requests are implemented in other
        methods:

        * :attr:`get_new_token_from_code`
        * :attr:`get_refreshed_token`
        * :attr:`get_scoped_app_token`
        * :attr:`get_new_token_from_code_with_pkce`

        :param kw:
            Keyword parameters passed to this method are interpreted as
            parameters to set in the body of the request.
        :returns:
            The JSON response from ``identity.pagerduty.com`` as a dictionary
            after amending via :attr:`amended_auth_response`. It should contain
            a key ``access_token`` with the new token and ``expiration_date``
            containing the date and time when the token will expire in ISO8601
            format.
        """
        return self.amended_auth_response(
            self.post(
                "/oauth/token",
                data=deepcopy(kw),
                headers={  # Overrides the default which would be `application/json`:
                    "Content-Type": "application/x-www-form-urlencoded"
                },
            )
        )

    def get_new_token_from_code(
        self, auth_code: str, scope: str, redirect_uri: str
    ) -> dict:
        """
        Exchange an authorization code granted by the user for an access token.

        :param auth_code:
            The authorization code received by the application at the redirect
            URI provided.
        :param scope:
            The scope of the authorization request.
        :redirect_uri:
            The redirect URI that was used in the authorization request.
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict
        """
        return self.get_new_token(
            grant_type="authorization_code",
            code=auth_code,
            scope=scope,
            redirect_uri=redirect_uri,
        )

    def get_refreshed_token(self, refresh_token: str) -> dict:
        """
        Obtain a new access token using a refresh token.

        :param refresh_token:
            The refresh token provided in the response of the original access
            token, i.e. in the dict returned by :attr:`get_new_token`
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict
        """
        return self.get_new_token(
            grant_type="refresh_token", refresh_token=refresh_token
        )

    def get_scoped_app_token(self, scope: str):
        """
        Obtain a scoped app token.

        This can be used to grant a server-side app a scoped non-user
        application token.

        :param scope:
            The scope of the authorization request.
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict
        """
        return self.get_new_token(grant_type="client_credentials", scope=scope)

    def generate_s256_pkce_params(self) -> Tuple[str, str]:
        """
        Generate PKCE parameters for OAuth authorization.

        Creates a code verifier and corresponding SHA256-hashed code challenge
        parameter for OAuth login with PKCE.

        :returns:
            A tuple containing ``(code_verifier, code_challenge)``. The code
            verifier must be temporarily stored so that it can be used in the
            final step of authorization, i.e. as the ``code_verifier`` argument
            to ``get_new_token_from_code_with_pkce``. The code challenge must
            be included in the query parameters of the initial authorize URL,
            i.e. set it as the the ``code_challenge`` argument to
            :attr:`get_pkce_authorize_url` to generate the URL.
        """
        code_verifier = (
            base64.urlsafe_b64encode(secrets.token_bytes(32))
            .decode("utf-8")
            .rstrip("=")
        )
        code_challenge = (
            base64.urlsafe_b64encode(
                hashlib.sha256(code_verifier.encode("utf-8")).digest()
            )
            .decode("utf-8")
            .rstrip("=")
        )
        return code_verifier, code_challenge

    def get_pkce_authorize_url(
        self, scope: str, redirect_uri: str, code_challenge: str
    ) -> str:
        """
        Generate an OAuth authorization URL with PKCE parameters ("Leg 1 of 3").

        :param scope:
            Scope of the OAuth grant requested
        :param redirect_uri:
            The redirect URI in the application to receive the code
        :param code_challenge:
            The code challenge generated from the code verifier, i.e. as
            returned by :attr:`generate_s256_pkce_params`.
        :returns:
            The formatted PKCE authorize URL
        """
        params = [
            ("client_id", self.client_id),
            ("redirect_uri", redirect_uri),
            ("response_type", "code"),
            ("scope", scope),
            ("code_challenge", code_challenge),
            ("code_challenge_method", "S256"),
        ]
        return self.url + "/oauth/authorize?" + urllib.parse.urlencode(params)

    def get_new_token_from_code_with_pkce(
        self, auth_code: str, scope: str, redirect_uri: str, code_verifier: str
    ) -> dict:
        """
        Exchange an authorization code for an access token using PKCE

        This is "Leg 3 of 3" in the PKCE authentication flow process.

        :param auth_code:
            The authorization code received by the application code via its
            inclusion in the redirect URI
        :param scope:
            The scope of the authorization request
        :param redirect_uri:
            The redirect URI that was used in the authorization request
        :param code_verifier:
            The code verifier used to generate the code challenge in Leg 1,
            i.e. as returned by :attr:`generate_s256_pkce_params`.
        :returns:
            The JSON response from ``identity.pagerduty.com``, containing a key
            ``access_token`` with the new token, as a dict.
        """
        return self.get_new_token(
            grant_type="authorization_code",
            code=auth_code,
            scope=scope,
            redirect_uri=redirect_uri,
            code_verifier=code_verifier,
        )

    def refresh_client(
        self,
        access_token: str,
        refresh_token: str,
        expiration_date: str,
        base_url: str = "https://api.pagerduty.com",
        **kw,
    ) -> Tuple[RestApiV2Client, Optional[dict]]:
        """
        Instantiate and return a :class:`pagerduty.RestApiV2Client` object

        Performs a token refresh if the current time is later than
        :attr:`early_refresh_buffer` seconds before the expiration date.

        :param access_token:
            The current REST API access token.
        :param refresh_token:
            The refresh token required to refresh the access token.
        :param expiration_date:
            The expiration date of the access token, formatted as an ISO8601
            datetime string, including the timezone suffix as a UTC offset. The
            value contained in the ``expiration_date`` key of the amended
            response dictionary returned by any of the "get token" methods
            should be usable as this parameter.
        :param base_url:
            The value to use for the ``url`` attribute of the client object.
        :param kw:
            Keyword arguments to pass to the constructor of the API client.
        :returns:
            A tuple containing a :class:`pagerduty.RestApiV2Client` object as
            its first element and the the amended OAuth response if a refresh
            was performed (and ``None`` otherwise) as its second element.
        """
        auth = None
        current_access_token = access_token
        if (
            datetime_to_relative_seconds(expiration_date)
            < self.early_refresh_buffer
        ):
            auth = self.get_refreshed_token(refresh_token)
            current_access_token = auth["access_token"]
        client_kw = deepcopy(kw)
        client_kw.update({"auth_type": "bearer"})
        client = RestApiV2Client(current_access_token, **client_kw)
        client.url = base_url
        return client, auth



from typing import Optional
import uuid

from .api_client import ApiClient
from .auth_method import AuthMethod
from .common import successful_response
from .errors import ServerHttpError


class McpApiClient(ApiClient):
    """
    Client class for the PagerDuty MCP API

    Usage example:

    .. code-block:: python

        # Import and use OAuthTokenAuthMethod instead of TokenAuthMethod to use
        # an application OAuth token:
        from pagerduty import (
            McpApiClient,
            TokenAuthMethod
        )

        # Instantiate:
        client = McpApiClient(TokenAuthMethod(API_KEY))

        # Call a method and get the result:
        result = client.call('tools/list')['result']
    """

    _url = "https://mcp.pagerduty.com"

    permitted_methods = ("POST",)

    def __init__(self, auth_method: AuthMethod, debug=False, **kw):
        super(McpApiClient, self).__init__(auth_method, debug=debug, **kw)
        self.headers.update({"Accept": "application/json, text/event-stream"})

    def call(
        self, method: str, params: Optional[dict] = None, req_id=None
    ) -> dict:
        """
        Make a JSON-RPC request to the MCP API.

        :param method:
            The JSON-RPC method to invoke.
        :param params:
            The parameters to send to the RPC method.
        :param req_id:
            A unique ID to send with the request. A random UUID will be used if
            this argument is unspecified; the ID will then be passed back in
            the response.
        :returns:
            The JSON-decoded response body; it will be a dictionary containing
            a "result" key with the response data.
        """
        if not req_id:
            req_id = str(uuid.uuid4())
        body = {
            "jsonrpc": "2.0",
            "id": req_id,
            "method": method,
        }
        if params:
            body["params"] = params
        response = successful_response(self.post("/mcp", json=body))
        response_body = response.json()
        if "result" not in response_body:
            raise ServerHttpError(
                "JSON-RPC response from PagerDuty did not include the "
                + 'expected "result" key.',
                response,
            )
        return response_body


__all__ = ["McpApiClient"]



from typing import List

from .rest_api_v2_base_client import CanonicalPath, RestApiV2BaseClient


CANONICAL_PATHS = ["/rules"]

ENTITY_WRAPPER_CONFIG = {
    # The /rules endpoints follow classic conventions; the wrapper can be inferred.
    #
    # This dictionary was intentionally created and left empty so that it is fewer steps
    # to support antipatterns if they get added to this API in the future.
}


class JiraServerIntegrationApiClient(RestApiV2BaseClient):
    """
    Client for the PagerDuty Jira Server Integration API.

    Inherits from :class:`pagerduty.RestApiV2BaseClient`.

    This client provides an abstraction layer for the `PagerDuty Jira Server
    Integration API
    <https://developer.pagerduty.com/api-reference/47fca619cb161-pager-duty-jira-server-integration-api>`_.

    Its documentation indicates that it only supports "Bearer" (OAuth-based)
    authentication and does not support the original token-style API
    authentication. For that reason, its constructor does not accept an
    "auth_type" argument, and it is assumed that the provided OAuth token was
    generated using an OAuth flow.

    :param access_token:
        OAuth access token (Bearer token) obtained via an OAuth flow
    :param jira_signature_token:
        Connected Jira instance signature token. This validates the connection
        between PagerDuty and a specific Jira instance.
    :param debug:
        Sets :attr:`pagerduty.ApiClient.print_debug`. Set to ``True`` to enable
        verbose command line output.
    """

    _url = "https://app.pagerduty.com/integration-jira-service"

    permitted_methods = ("GET", "POST", "PUT", "DELETE")

    def __init__(
        self,
        access_token: str,
        jira_signature_token: str,
        debug: bool = False,
        **kw,
    ):
        super(JiraServerIntegrationApiClient, self).__init__(
            access_token, auth_type="bearer", debug=debug, **kw
        )

        self.jira_signature_token = jira_signature_token
        self.headers.update(
            {
                "Accept": "application/json",
                "x-pagerduty-jira-signature": self.jira_signature_token,
            }
        )

    @property
    def canonical_paths(self) -> List[CanonicalPath]:
        return CANONICAL_PATHS

    @property
    def entity_wrapper_config(self) -> dict:
        return ENTITY_WRAPPER_CONFIG
